// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#ifndef KERNEL_LIB_KERNEL_LEARNING_ALIGNMENT_H_
#define KERNEL_LIB_KERNEL_LEARNING_ALIGNMENT_H_

#include <iostream>
#include <math.h>
#include <vector>

#include "kernel/lib/quadprog.h"
#include "kernel/lib/io-utilities.h"
#include "kernel/lib/dense-vector.h"
#include "kernel/lib/dense-matrix.h"
#include "kernel/lib/kernel-learning-linearl1.h"

namespace kernel {

template <typename T> 
double Alignment(const DenseMatrix<T> &kernel_in, 
                 const DenseVector<T> &labels_in,
                 bool center_kernels=true,
                 bool center_labels=true) {

  // Center data and labels
  DenseMatrix<T> kernel = kernel_in;
  if (center_kernels)
    CenterKernel(&kernel);
  DenseVector<T> labels = labels_in;
  if (center_labels)
    Sum(-1 * Mean(labels), DenseVector<double>(labels.Size(), 1.0), &labels);

  // <K, yy'>_F
  double Kyy = 0;
  // <K, K>_F
  double KK = 0;

  for (int i = 0; i < kernel.NumRow(); ++i)
    for (int j = 0; j < kernel.NumColumn(); ++j) {
      Kyy += kernel.Element(i,j) * labels.Element(i) * labels.Element(j);
      KK += pow(kernel.Element(i,j), 2);
    }

  KK = sqrt(KK);

  // <yy', yy'>_F
  double yy = 0;

  for (int i = 0; i < labels.Size(); ++i)
    yy += pow(labels.Element(i), 2);

  VLOG(1) << "<K, yy>: " << Kyy << " ||K||_F: " << KK << " ||yy||_F: " << yy;

  return Kyy / (KK * yy);
}

// Generate a weighting such that each kernel is weighted proportional
// to its alignment
template <typename T>
void MakeAlignmentWeights(const vector< DenseMatrix<T> > &kernels,
                     const DenseVector<T> &labels,
                     double norm,
                     double center_kernels,
                     double center_labels,
                     DenseVector<T> *alignment_weights) {

  alignment_weights->Resize(kernels.size(), 0);

  for (int i = 0; i < kernels.size(); ++i) {
    alignment_weights->SetElement(i, 
                 Alignment(kernels[i], labels, center_labels, center_kernels));
  }

  //Normalize weights, L1(weights) = norm
  double sum = Sum(*alignment_weights);
  for (int i = 0; i < kernels.size(); ++i) {
    alignment_weights->SetElement(i, 
                  alignment_weights->Element(i) / sum * norm);
  }
}

template <typename T>
void MakeFullAlignmentWeights(const vector< DenseMatrix<T> > &kernels_in,
                     const DenseVector<T> &labels_in,
                     double norm,
                     bool center_kernels,
                     bool center_labels,
                     DenseVector<T> *alignment_weights) {

  // Center data and labels
  vector< DenseMatrix<T> > kernels(kernels_in.size());
  for (size_t i = 0; i < kernels_in.size(); ++i) {
    kernels[i] = kernels_in[i];
    if (center_kernels)
      CenterKernel(&kernels[i]);
  }
  DenseVector<T> labels = labels_in;
  if (center_labels)
    Sum(-1 * Mean(labels), DenseVector<double>(labels.Size(), 1.0), &labels);

  int p = kernels.size();
  DenseMatrix<T> M(p, p, 0.0); // M_ij = <K_i, K_j>_F
  DenseVector<T> a(p, 0.0);    // a_i = <K_i, yy^t>_F = y^t * K_i * y

  DenseMatrix<T> H(p, p, 0.0); // H = 1/4 * Minv 
  DenseVector<T> f(p, 0.0);    // f = Minv * a
  DenseVector<T> beta_pos(p, 0.0);
  DenseVector<T> weights_pos(p, 0.0);
  DenseVector<T> temp(p, 0.0);

  // Compute M, symmetric
  double val;
  for (int i = 0; i < p; ++i)
    for (int j = 0; j < i; ++j) {
      val = FrobeniusProductSym(kernels[i], kernels[j]);
      M.SetElement(i, j, val);
      M.SetElement(j, i, val);
    }

  for (int i = 0; i < p; ++i)
    M.SetElement(i, i, FrobeniusProductSym(kernels[i], kernels[i]));

  // Compute a
  DenseVector<double> Ky;
  for (int i = 0; i < p; ++i) {
    Product(kernels[i], labels, &Ky);
    a.SetElement(i, DotProduct(Ky, labels));
  }

  DenseMatrix<double> Minv(p, p, 0);
  PDSInverse(M, &Minv);

/*
  LOG(DEBUG) << "M:";
  PrintDenseMatrix(M, LOG(DEBUG));
  PrintDenseMatrix(Minv, LOG(DEBUG));
*/
  DenseMatrix<double> evec;
  DenseVector<double> eval;
  SymEigen(Minv, &evec, &eval);
/*
  LOG(debug) << "eig:";
  PrintVector(eval, LOG(INFO));
*/

  Product(Minv, a, &f);
  ScalarProduct(0.5, Minv, &H); // opt includes 0.5 multiplier

/*
  LOG(DEBUG) << "a:";
  PrintVector(a, LOG(DEBUG));
 */ 

  // scale objective for better convergence
  double scale = Norm(f);
	scale = scale / f.Size();
  ScalarProduct(1/scale, &H);
  ScalarProduct(1/scale, &f);

/*
  LOG(DEBUG) << "f:";
  PrintVector(f, LOG(DEBUG));
 
  LOG(DEBUG) << "H:";
  PrintDenseMatrix(H, LOG(DEBUG));
*/

  double pos_obj;
  if (H.NumRow() > 500) {
    pos_obj = QP_LibLinear(H, f, &beta_pos);
  } else {
    pos_obj = QP_ActiveSet(H, f, &beta_pos); // Slower but more accurate
  }


  for (size_t i = 0; i < beta_pos.Size(); ++i) 
    if (beta_pos.Element(i) < -1e-9)
      LOG(WARNING) << "MakeFullAlignmentWeights: Invalid QP solution";

  // weights_pos = Minv * (0.5 * beta_pos + a)
  Sum(0.5, beta_pos, a, &temp);
  Product(Minv, temp, &weights_pos);
  if (Norm(weights_pos) < 1e-8) {
    weights_pos.Resize(0, 0); // non-useful solution
  } else  
    ScalarProduct(norm / Norm(weights_pos), &weights_pos);

  for (size_t i = 0; i < weights_pos.Size(); ++i)  {
    if (weights_pos.Element(i) < -1e-6)
      LOG(WARNING) 
        << "MakeFullAlignmentWeights: truncating negative combination weights";
    if (weights_pos.Element(i) < 0)
      weights_pos.SetElement(i, 0);
  }

  *alignment_weights = weights_pos;
}

template <typename T>
void MakeOnestageAlignmentWeights(const vector< DenseMatrix<T> > &kernels_in,
                     const DenseVector<T> &labels_in,
                     double lambda, 
                     double g0, double g1, double g2, //tuneable parameters
                     bool center_kernels,
                     bool center_labels,
                     int max_iter,
                     DenseVector<T> *mu) {

  // Center data and labels
  vector< DenseMatrix<T> > kernels(kernels_in.size());
  for (size_t i = 0; i < kernels_in.size(); ++i) {
    kernels[i] = kernels_in[i];
    if (center_kernels)
      CenterKernel(&kernels[i]);
  }
  DenseVector<T> labels = labels_in;
  if (center_labels)
    Sum(-1 * Mean(labels), DenseVector<double>(labels.Size(), 1.0), &labels);

  int p = kernels.size();
  int m = labels.Size();
  DenseMatrix<T> M(p, p, 0.0); // M_ij = <K_i, K_j>_F
  DenseVector<T> a(p, 0.0); // a_i = <K_i, yy^t>_F = y^t * K_i * y

  double eps = 1e-5;
  double eta = 1;

  // Compute M, symmetric
  double val;
  for (int i = 0; i < p; ++i)
    for (int j = 0; j < i; ++j) {
      val = FrobeniusProductSym(kernels[i], kernels[j]);
      M.SetElement(i, j, val);
      M.SetElement(j, i, val);
    }

  for (int i = 0; i < p; ++i)
    M.SetElement(i, i, FrobeniusProductSym(kernels[i], kernels[i]));

  // Compute a
  DenseVector<double> Ky;
  for (int i = 0; i < p; ++i) {
    Product(kernels[i], labels, &Ky);
    a.SetElement(i, DotProduct(Ky, labels));
  }

  double diff = 1e100;
  double obj = 1e100;
  int iter = 1;
  // initialize mu to maximum alignment solution
  MakeFullAlignmentWeights(kernels_in, labels_in, 1.0, center_kernels,
                           center_labels, mu);
  /*
  mu->Resize(p, 0.0);
  for (int i = 0; i < mu->Size(); ++i)
    mu->SetElement(i, 1.0);
  */

  DenseVector<T> mu_raw;
  mu_raw = *mu;

  DenseMatrix<T> K_mu(p, p, 0.0);
  LinearCombination(kernels, *mu, &K_mu);

  DenseVector<T> alpha(m, 0.0);
  KRRDualTrain(K_mu, labels, lambda, &alpha);

  DenseVector<T> beta(p, 0.0);
  Product(M, *mu, &beta);

  DenseVector<T> mu_old, mu_raw_old, alpha_old, beta_old;
  DenseMatrix<T> K_mu_old;
  double obj_old;
  DenseVector<T> grad(p, 0.0);

  DenseVector<T> tempv;
  DenseMatrix<T> tempm;

  while (diff > eps && iter < max_iter) {
    mu_old = *mu;
    mu_raw_old = mu_raw;
    obj_old = obj;
    K_mu_old = K_mu;
    alpha_old = alpha;
    beta_old = beta;

    for (int i = 0; i < grad.Size(); ++i) {
      Product(kernels[i], alpha, &tempv);
      // -alpha' * K[i] * alpha 
      grad.SetElement(i, -DotProduct(alpha, tempv));
      
      grad.SetElement(i, grad.Element(i) - g0 * a.Element(i));

      grad.SetElement(i, grad.Element(i) + 2 * g1 * mu->Element(i));

      grad.SetElement(i, grad.Element(i) + 2 * g2 * beta.Element(i));
    }

    ScalarProduct(-eta, &grad);
    Sum(mu_old, grad, mu);
    mu_raw = *mu;
    for (int i = 0; i < mu->Size(); ++i)
      if (mu->Element(i) < 0) mu->SetElement(i, 0.0);

    LinearCombination(kernels, *mu, &K_mu);
    // alpha = inv(K_mu + lambda * eye(m)) * y;
    KRRDualTrain(K_mu, labels, lambda, &alpha);
    Product(M, *mu, &beta);
    
    ScalarProduct(-1.0, mu_raw_old, &tempv);
    Sum(mu_raw, &tempv);
    diff = Norm(tempv) / max( max( Norm(mu_raw), Norm(mu_raw_old) ), 1e-19);

    ScalarProduct(g2, M, &tempm);
    for (int i = 0; i < M.NumRow(); ++i) 
      tempm.SetElement(i, i, tempm.Element(i,i) + g1);
    Product(tempm, *mu, &tempv);
    obj = DotProduct(labels, alpha) - g0 * DotProduct(*mu, a)
          // mu' * (g1 * I + g2 * M) * mu
          + DotProduct(tempv, *mu);

    LOG(INFO) << "iter: " << iter << " eta: " << eta
              << " obj: " << obj << " diff: " << diff;

    if (obj_old < obj) {
      eta = eta * 0.5;
      *mu = mu_old;
      mu_raw = mu_raw_old;
      obj = obj_old;
      K_mu = K_mu_old;
      alpha = alpha_old;
      beta = beta_old;
    }

    ++iter;
  }
}

/*
// Center the data K in the kernel space, according to the statistics
// (x_bar) in G.
//   K_centered(x_i,x_j)
//      = K(x_i,x_j) - K(x_i, x_bar) - K(x_j, x_bar) + K(x_bar, x_bar).
template <typename U, typename T>
void CenterKernel(const DenseVector<U> &train_points, DenseMatrix<T> *K) {
  DenseVector<U> all_points;
  MakeRange(0, K->NumColumn(), &all_points); 

  size_t m = train_points.Size();
  size_t n = all_points.Size();
  double Gbar = (1.0 / pow(m, 2.0)) * Sum(*K, train_points, train_points);
  double GKrow[n];
  DenseVector<U> row(1,0);
  for (size_t i = 0; i < n; ++i) {
    row.SetElement(0, i);
    GKrow[i] = (1.0 / m) * Sum(*K, row, train_points);
  }

  double temp;
  for (size_t i = 0; i < n; ++i)
    for (size_t j = 0; j < i; ++j) {
      temp = K->Element(i, j) - GKrow[i] - GKrow[j] + Gbar;
      K->SetElement(i, j, temp);
      K->SetElement(j, i, temp);
    }
  for (size_t i = 0; i < n; ++i)
    K->SetElement(i, i, K->Element(i,i) - 2 * GKrow[i] + Gbar);
}

// Center the data in the kernel space.
//   K_centered(x_i,x_j)
//      = K(x_i,x_j) - K(x_i, x_bar) - K(x_j, x_bar) + K(x_bar, x_bar).
template <typename T>
void CenterKernel(DenseMatrix<T> *K) {
  DenseVector<size_t> all_points;
  MakeRange(0, K->NumColumn(), &all_points); 
  
  CenterKernel(all_points, K);
}
*/

/*
// TODO: finish this writing this function once we have a faster
//       QP solver.
// Return feature weights (i.e. sqrts of rank-1 kernel weights)
// that maximize the final kernel alignment.
// Note: centers the labels and data as part of QP setup
template <class M, typename T>
void MakeFullAlignmentWeightsRank1(const M &features,
                     const DenseVector<T> &labels_in,
                     double norm,
                     DenseVector<T> *alignment_weights) {
  
  size_t m = features.NumRow();
  size_t n = features.NumColumn();
  DenseMatrix<T> H, U;
  DenseVector<T> a, a_neg, u, v_pos, v_neg;
  typename M::TransposeMatrix features_t;
  features.Transpose(&features_t);

  // H = Xc * Xc, where Xc = X - 1/m ones * ones' X (i.e. centered features)
  LOG(INFO) << "Setting up quadratic constants";
  Product(features_t, features, &H);
  Product(features_t, DenseVector<T> (m, 1.0), &u);
  OuterProduct(u, u, &U);
  Sum(-1.0 / m, U, 1.0, &H);

  // a = (Xc' * yc).^2
  LOG(INFO) << "Setting up linear constants";
  DenseVector<T> labels = labels_in; // center labels
  Sum(-1 * Mean(labels), DenseVector<T>(labels.Size(), 1.0), &labels);
  Product(features_t, labels, &a);
  Sum(-1 * DotProduct(labels, DenseVector<T>(m, 1.0)) / m, // -<y, 1> / m
      u, &a);
  ScalarProduct(-1.0, a, &a_neg);
      
  // Solve min_{v >= 0} 0.5 *  v' * M * v +/- v' * a
  LOG(INFO) << "Solving positive case";
  SimpleQP(H, a, &v_pos);

  LOG(INFO) << "Solving negative case";
  SimpleQP(H, a_neg, &v_neg);

  PrintVector(v_pos, LOG(INFO));
  PrintVector(v_neg, LOG(INFO));

  //TODO: find neg/pos weights, see which one does better
}
*/

/*
// Generate weighting \Mu s.t. Alignment(\sum_i \mu_i \K_i) is maximal, 
// subject to \Mu \succeq 0.
template <typename T>
void MakeFullAlignmentWeights(const vector< DenseMatrix<T> > &kernels,
                     const DenseVector<T> &labels,
                     double norm,
                     size_t max_iter,
                     double stop_tol,
                     DenseVector<T> *alignment_weights) {
  int p = kernels.size();
  DenseMatrix<T> M(p, p, 0.0); // M_ij = <K_i, K_j>_F
  DenseVector<T> a(p, 0.0); // a_i = <K_i, yy^t>_F = y^t * K_i * y

  DenseMatrix<T> H(p, p, 0.0); // H = 1/4 * Minv 
  DenseVector<T> f(p, 0.0); // f = Minv * a
  DenseVector<T> beta_pos(p, 0.0);
  DenseVector<T> beta_neg(p, 0.0);
  DenseVector<T> weights_pos(p, 0.0);
  DenseVector<T> weights_neg(p, 0.0);
  DenseVector<T> temp(p, 0.0);

  // Compute M, symmetric
  double val;
  for (int i = 0; i < p; ++i)
    for (int j = 0; j < i; ++j) {
      val = FrobeniusProductSym(kernels[i], kernels[j]);
      M.SetElement(i, j, val);
      M.SetElement(j, i, val);
    }

  for (int i = 0; i < p; ++i)
    M.SetElement(i, i, FrobeniusProductSym(kernels[i], kernels[i]));

  // Compute a
  DenseVector<double> Ky;
  for (int i = 0; i < p; ++i) {
    Product(kernels[i], labels, &Ky);
    a.SetElement(i, DotProduct(Ky, labels));
  }

  DenseMatrix<double> Minv(p, p, 0);
  PDSInverse(M, &Minv);

  // Solve for both positive and negative case
  Product(Minv, a, &f);
  ScalarProduct(0.25, Minv, &H);
	LOG(INFO) << "Solving positive case QP...";
	SolveAlignmentQP(H, f, max_iter, stop_tol, &beta_pos);
	// weights_pos = Minv * (0.5 * beta_pos + a)
	Sum(0.5, beta_pos, a, &temp);
	Product(Minv, temp, &weights_pos);
	ScalarProduct(1 / Norm(weights_pos), &weights_pos);

	ScalarProduct(-1.0, &f);
	LOG(INFO) << "Solving negative case QP...";
	SolveAlignmentQP(H, f, max_iter, stop_tol, &beta_neg);
	// weights_neg = Minv * (0.5 * beta_neg - a)
	Sum(-0.5, beta_neg, a, &temp);
	ScalarProduct(-1.0, &temp);
	Product(Minv, temp, &weights_neg);
	ScalarProduct(1 / Norm(weights_neg), &weights_neg);

  LOG(INFO) << "pos:";
	PrintVector(weights_pos, LOG(INFO));
  LOG(INFO) << "neg:";
	PrintVector(weights_neg, LOG(INFO));

	//FIXME: convergence of gradient descent not sufficient
  
}


// obj = beta' * H * beta + f' * beta
template <typename T>
double AlignmentQPObj(const DenseMatrix<T> &H, const DenseVector<T> &f, 
                      const DenseVector<T> &beta) {
  DenseVector<T> temp(f.Size(), 0);
	Product(H, beta, &temp);
	return DotProduct(temp, beta) + DotProduct(f, beta);
}

// Solve:
//  min_{beta >= 0} beta' * H beta + f' * beta
template <typename T>
void SolveAlignmentQP(const DenseMatrix<T> &H, const DenseVector<T> &f, 
                      size_t max_iter, double stop_tol,
                      DenseVector<T> *sol) {
  CHECK(H.NumRow() == H.NumColumn());
  CHECK(H.NumRow() == f.Size());
  double STEP_TOL = 1e-8;

  double diff = 1e100; //infinity
  size_t iter = 0;
  double step = 1.0, old_obj, obj;
  DenseVector<T> beta(f.Size(), 1);
  DenseVector<T> beta_cand(f.Size(), 1);
  DenseVector<T> grad(f.Size(), 0);

  old_obj = AlignmentQPObj(H, f, beta);
  obj = old_obj;
  while ((abs(diff) > stop_tol) && (max_iter > iter) && (step > STEP_TOL)) {
    VLOG(2) << "iter: " << iter << " obj: " << obj 
		        << " gap: " << diff << " step: " << step;
    
    // grad = 2 * H beta + f
    grad = f;
    Product(2.0, H, beta, 1.0, &grad);

		diff = -1;
		while ((diff < 0) && (step > STEP_TOL)) {
      // beta_cand = beta - step * grad 
      Sum(-1.0 * step, grad, beta, &beta_cand);
			
      // beta_cand = max(beta_cand, 0)
      for (size_t i = 0; i < beta_cand.Size(); ++i)
  		  if (beta_cand.Element(i) < 0)
	  		  beta_cand.SetElement(i, 0);

      obj = AlignmentQPObj(H, f, beta_cand);

      diff = (old_obj - obj) / abs(old_obj);
			VLOG(3) << "diff: " << diff << " step : " << step;

      if (diff < 0)
        step = step / 2;
    }
		step = step * 2;

		beta = beta_cand;
    old_obj = obj;

    ++iter;
  } 

  LOG(INFO) << "num_iter: " << iter << " obj: " << obj 
		        << " gap: " << diff << " step: " << step;
	*sol = beta;
}
*/

/*
// Generate weighting \Mu s.t. Alignment(\sum_i \mu_i \K_i) is maximal, 
// subject to \Mu \succeq 0.
template <typename T>
void MakeFullAlignmentWeightsOld(const vector< DenseMatrix<T> > &kernels,
                     const DenseVector<T> &labels,
                     double norm,
                     bool clip,
                     DenseVector<T> *alignment_weights) {
  int p = kernels.size();
  DenseMatrix<T> M(p, p, 0.0); // M_ij = <K_i, K_j>_F
  DenseVector<T> a(p, 0.0); // a_i = <K_i, yy^t>_F = y^t * K_i * y

  // Compute M, symmetric
  double val;
  for (int i = 0; i < p; ++i)
    for (int j = 0; j < i; ++j) {
      val = FrobeniusProductSym(kernels[i], kernels[j]);
      M.SetElement(i, j, val);
      M.SetElement(j, i, val);
    }

  for (int i = 0; i < p; ++i)
    M.SetElement(i, i, FrobeniusProductSym(kernels[i], kernels[i]));

  // Compute a
  DenseVector<double> Ky;
  for (int i = 0; i < p; ++i) {
    Product(kernels[i], labels, &Ky);
    a.SetElement(i, DotProduct(Ky, labels));
  }

  DenseMatrix<double> Minv(p, p, 0);
  PDSInverse(M, &Minv);

  Product(Minv, a, alignment_weights);

  //choose pos coords and normalize
  if (clip)
    ClipAndNormalize(*alignment_weights, M, a, labels, norm, alignment_weights);

}

template<typename T> 
void ClipAndNormalize(const DenseVector<T> &weights, 
                      const DenseMatrix<T> &M,
                      const DenseVector<T> &a,
                      const DenseVector<T> &labels,
                      double norm,
                      DenseVector<T> *clipped_weights) {
  clipped_weights->Resize(weights.Size(), 0);
  DenseVector<T> w_pos = DenseVector<T>(weights.Size(), 0);
  DenseVector<T> w_neg = DenseVector<T>(weights.Size(), 0);

  for (size_t i = 0; i < weights.Size(); ++i) {
    weights.Element(i) < 0 ? 
      w_pos.SetElement(i, 0) : w_pos.SetElement(i, weights.Element(i));
    weights.Element(i) > 0 ? 
      w_neg.SetElement(i, 0) : w_neg.SetElement(i, -1 * weights.Element(i));
  }

  // normalize vectors
  double norm_pos = Norm(w_pos);
  ScalarProduct(norm / norm_pos, &w_pos);
  double norm_neg = Norm(w_neg);
  ScalarProduct(norm / norm_neg, &w_neg);

  // which sign provides best 
  DenseVector<T> temp(weights.Size());
  double obj_pos = pow(DotProduct(w_pos, a), 2);
  Product(M, w_pos, &temp);
  obj_pos = obj_pos / DotProduct(temp, w_pos);
  
  double obj_neg = pow(DotProduct(w_neg, a), 2);
  Product(M, w_neg, &temp);
  obj_neg = obj_neg / DotProduct(temp, w_neg);
    
  if (obj_pos > obj_neg)
    *clipped_weights = w_pos;
  else
    *clipped_weights = w_neg;
}
*/

} //namespace kernel

#endif  // KERNEL_LIB_KERNEL_LEARNING_ALIGNMENT_H_
