// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)
//
// This file contains wrapper functions for using QuadProg++.

#ifndef KERNEL_LIB_QUADPROG_H_
#define KERNEL_LIB_QUADPROG_H_

#include <kernel/lib/liblinear.h>

namespace kernel {

// Solve simple QP:
// min 0.5 * x' * H * x + x * f
// s.t. x >= 0;
//
// Return the objective value.
//
// Solved via projected gradient descent:
// Works well sometimes, but can be inexact and sometimes has
// very slow converegence.
template <typename T>
double SimpleQP_GD(const DenseMatrix<T> &H, const DenseVector<T> &f,
                DenseVector<T> *sol) {
  CHECK(H.NumRow() == H.NumColumn());
  CHECK(H.NumRow() == f.Size());
  size_t p = f.Size();	

  size_t MAXITER = 100000;
  double stepsize = 1;
  double obj = 1e300;
  double obj_old, diff;

  DenseVector<T> x(p, 1.0);
  DenseVector<T> x_old(p, 0.0);
  DenseVector<T> grad(p, 1.0);
  DenseVector<T> Hx(p, 0.0);
  DenseVector<T> temp(p, 0.0);
  DenseMatrix<T> Hinv(p, p, 0.0);

  int iter;
  for(iter = 1; iter < MAXITER; iter++) {
    x_old = x;
    Product(H, x_old, &Hx);
    Sum(Hx, f, &grad);
    Sum(-stepsize, grad, x_old, &x);     // Gradient step

//    PDSInverse(H, &Hinv);              
//    Product(Hinv, grad, &temp);
//    Sum(-stepsize, temp, x_old, &x);   // Hessian step

    // enforce x >= 0
    for(size_t i = 0; i < x.Size(); ++i)
      if(x.Element(i) < 0)
        x.SetElement(i, 0);

    //PrintVector(x, cout); //DEBUG

    obj_old = obj;
    Product(H, x, &Hx);
    obj = 0.5 * DotProduct(Hx, x);
    obj += DotProduct(f, x);
    diff = obj_old - obj;

    // if no improvement in objective
    // or if step forces x = 0
    if(diff < 0 || Sum(x) == 0) {
      obj = obj_old;
      x = x_old;
      stepsize = stepsize / 2;
      iter -= 1;
    } else {
      stepsize *= 2;
    }

    VLOG(2) << ("iter: %d obj: %g diff: %g stepsize: %g gnorm: %g\n", 
           iter, obj, diff, stepsize, Norm(grad));

    if((stepsize < 1e-12) || (diff > 0 && diff < Norm(x) * 1e-18))
      break;
  }
  if((stepsize < 1e-12) || (iter >= MAXITER))
    LOG(WARNING) << "QP has not converged";

  *sol = x;

  return obj;
}


template <typename T>
double QP_LibLinear(const DenseMatrix<T> &H_in,
               const DenseVector<T> &f_in,
               DenseVector<T> *x) {

  DenseMatrix<T> H(H_in);
  DenseVector<T> f(f_in);

  DenseMatrix<T> M;

  Cholesky(H, &M); // create a dataset from the gram matrix
  M.Transpose(&M);

  LOG(DEBUG) << "M: " << endl;
  PrintDenseMatrix(M, LOG(DEBUG));

  LOG(DEBUG) << "H_outside:" << endl;
  DenseMatrix<T> Mt; M.Transpose(&Mt);
  Product(M, Mt, &H);
  PrintDenseMatrix(H, LOG(DEBUG));

  DenseVector<T> y(H.NumRow(), 1.0);

  LibLinearTrain(M, y, 1e100, f, x);

  double obj = 0;
  DenseVector<T> temp;
  Product(H, *x, &temp);
  obj += 0.5 * DotProduct(*x, temp);
  obj += DotProduct(*x, f);

  return obj;
}

// Solve simple QP:
// min 0.5 * x' * H * x + x * f
// s.t. x >= 0;
//
// Return the objective value.

// OpenKernel wrapper function.
// Solves the l1-loss svm problem, given the kernel matrix.
// // INPUT:
// kernel - a Dense matrix containing the kernel matrix.
// labels - a DenseVector of labels for each instance,
// reg_param - the SVM regularization parameter "C", which upper bounds
//   the size of each dual variable.
// OUTPUT:
// dual_vars - hold the optimal dual variables
template <typename T>
double QP_CoordDescent(const DenseMatrix<T> &Hconst,
               const DenseVector<T> &fconst,
               DenseVector<T> *x) {

  DenseVector<T> f(fconst);
  DenseMatrix<T> H(Hconst);
  CHECK(f.Size() == H.NumRow());
  CHECK(f.Size() == H.NumColumn());

  size_t max_iter = 200000;
  size_t iter = 0;
  double eps = 1e-12;
  double diff;
  double g;
  double pg;
  int m = f.Size();

  DenseVector<T> x_old;
  x->Clear();
  x->Resize(f.Size(), 0);

  while (iter < max_iter) {
    x_old = *x;

    for (int i = 0; i < m; ++i) {
    
      // gradient of ith coordinate
      g = 0;
      for (int j = 0; j < m; ++j) {
        g += x->Element(j) * H.Element(i,j);
      }
      g += f.Element(i);

      // projected gradited of ith coordinate
      if (x->Element(i) < 1e-12)
        pg = min(g, 0.0);
      else
        pg = g;

      if (abs(pg) > 1e-12)
        x->SetElement(i, 
          max(x->Element(i) - g / H.Element(i,i), 0.0));
    }

    diff = 0;
    for (int i = 0; i < x_old.Size(); ++i)
      diff += pow(x->Element(i) - x_old.Element(i), 2);
    diff = sqrt(diff) / x_old.Size();

    if (diff < eps)
      break;

    ++iter;
  }
  LOG(INFO) << "Optimization completed, num iter: " << iter;
  if (iter >= max_iter)
    LOG(WARNING) << "Optimization may not have converged.";

  double obj = 0;
  DenseVector<T> temp;
  Product(H, *x, &temp);
  obj += 0.5 * DotProduct(*x, temp);
  obj += DotProduct(*x, f);

  return obj;
}

// Translate binary vector into a list of active indices.
template <typename T, typename I>
void GetActiveIndex(const DenseVector<T> &active, 
                    DenseVector<I> *indices) {
  indices->Resize(0, 0);
  int j = 0;

  for (int i = 0; i < active.Size(); ++i) {
    if (active.Element(i)) {
      indices->AddElement(j, i);
      ++j;
    }
  }

}

// Solve simple QP:
// min 0.5 * x' * H * x + x * f
// s.t. x >= 0;
//
// Return the objective value.
//
// Accurate, but a slow method for large matrices H (e.g. > 500 rows).
template <typename T>
double QP_ActiveSet(const DenseMatrix<T> &H,
                    const DenseVector<T> &f,
                    DenseVector<T> *x) {

  size_t MAX_ITER = 10 * f.Size();
  size_t m = f.Size();
  double EPS = 1e-12;
  T x_min;
  T gradient_min;
  int x_i;
  int gradient_i;


  DenseMatrix<T> H_active;
  DenseVector<T> f_active;
  DenseVector<T> x_active;
  DenseVector<T> gradient;

  // active(i) = 1, if ith coord is active
  DenseVector<int> active(m, 1);
  DenseVector<int> active_old; 
  DenseVector<int> active_idx;


  size_t iter = 0;
  for (; iter < MAX_ITER; ++iter) {
    x_min = 0.0; x_i = -1;
    gradient_min = 0.0; gradient_i = -1;
    active_old = active;
    GetActiveIndex(active, &active_idx);
    SubMatrixCopy(H, active_idx, active_idx, &H_active);
    SubsetCopy(f, active_idx, &f_active);
    ScalarProduct(-1.0, &f_active);

    SolveLinearEqPDS(H_active, f_active, &x_active);

    x->Clear();
    x->Resize(m, 0.0);
    for (int i = 0; i < active_idx.Size(); ++i)
      x->SetElement(active_idx.Element(i), x_active.Element(i));

    Product(H, *x, &gradient); // gradientient computed with all coords
    Sum(gradient, f, &gradient);

    // check for negative valued active variables
    for (int i = 0; i < x->Size(); ++i) {
      if (x->Element(i) < x_min) {
        x_min = x->Element(i);
        x_i = i;
      }
    }

    if (x_min < -EPS)
      active.SetElement(x_i, 0);
    else
      x_i = -1;


    // check for negative graidents at non-active variables
    for (int i = 0; i < gradient.Size(); ++i) {
      if (gradient.Element(i) < gradient_min) {
        gradient_min = gradient.Element(i);
        gradient_i = i;
      }
    }

    if ((gradient_min < -EPS)         // is the gradient neg?
        && (active.Element(gradient_i) == 0)  // is this a non-active coord?
        && (gradient_i != x_i)) {   // did we just flip this coord?
      active.SetElement(gradient_i, 1);
    }

    if (Equal(active, active_old)) {
      LOG(INFO) << "Optimal solution found";
      break;
    }

  }

  if (iter >= MAX_ITER)
    LOG(WARNING) << "Maximum iteration reached, optimum not found.";

  double obj = 0;
  DenseVector<T> temp;
  Product(H, *x, &temp);
  obj += 0.5 * DotProduct(*x, temp);
  obj += DotProduct(*x, f);

  return obj;
}

} // namepace kernel

#endif  // KERNEL_LIB_QUADPROG_H_
